package Net::Async::Graphite::JSONCodec;

our $VERSION = '0.0_1';

use v5.14;
use strictures 2;
use Moo::Role;
use Carp;

use JSON ();
use namespace::clean;

=item _json (read-only)

A L<JSON> codec object.

=cut

has _json => is => lazy => init_arg => undef, builder => sub {
  JSON->new->utf8->allow_blessed(1)
};

1;
